import pytest
from devtools import debug

FRACTAL_DEFAULT_GROUP_NAME = "All"

PWD_USER = "1234"
PROJECT_DIR_USER = "/user-project-dir"


def test_register_as_user(invoke, caplog):
    with pytest.raises(SystemExit):
        invoke("user register aaa bbb /ccc")
    debug(caplog.text)
    assert "403" in caplog.text


@pytest.mark.parametrize("is_superuser", [True, False])
def test_register_as_superuser(
    invoke_as_superuser, is_superuser: bool, new_name
):
    EMAIL_USER = f"{new_name()}@example.org"
    if is_superuser:
        res = invoke_as_superuser(
            f"user register {EMAIL_USER} {PWD_USER} {PROJECT_DIR_USER} "
            "--superuser"
        )
        debug(res.data)
        assert res.retcode == 0
        assert res.data["is_superuser"]
    else:
        res = invoke_as_superuser(
            f"user register {EMAIL_USER} {PWD_USER} {PROJECT_DIR_USER}"
        )
        debug(res.data)
        assert res.retcode == 0
        assert not res.data["is_superuser"]
    assert res.data["email"] == EMAIL_USER

    # Test that new user is verified (note: for the moment we don't expose the
    # possibility of registering a non-verified user)
    assert res.data["is_verified"]


def test_register_as_superuser_with_batch(invoke_as_superuser, new_name):
    EMAIL_USER = f"{new_name()}@example.org"
    # Register a user with the --batch flag
    res = invoke_as_superuser(
        f"--batch user register {EMAIL_USER} {PWD_USER} {PROJECT_DIR_USER}"
    )
    user_id = res.data
    debug(user_id)
    assert res.retcode == 0
    # Check that the user exists
    res = invoke_as_superuser(f"user show {user_id}")
    debug(res.data)
    assert res.data["email"] == EMAIL_USER
    assert res.retcode == 0


def test_list_as_user(invoke, caplog):
    with pytest.raises(SystemExit):
        invoke("user list")
    debug(caplog.text)
    assert "403" in caplog.text


def test_list_as_superuser(invoke_as_superuser, superuser, tester):
    res = invoke_as_superuser("user list")
    debug(res.data)
    assert res.retcode == 0
    list_emails = [user["email"] for user in res.data]
    debug(list_emails)
    assert superuser["email"] in list_emails
    assert tester["email"] in list_emails


def test_show_as_user(invoke, invoke_as_superuser, caplog, new_name):
    EMAIL_USER = f"{new_name()}@example.org"
    # Register a new user
    res = invoke_as_superuser(
        f"user register {EMAIL_USER} {PWD_USER} {PROJECT_DIR_USER}"
    )
    user_id = res.data["id"]
    # Call fractal user show
    with pytest.raises(SystemExit):
        invoke(f"user show {user_id}")
    debug(caplog.text)
    assert "403" in caplog.text


def test_show_as_superuser(invoke_as_superuser, new_name):
    EMAIL_USER = f"{new_name()}@example.org"
    # Register a new user
    res = invoke_as_superuser(
        f"user register {EMAIL_USER} {PWD_USER} {PROJECT_DIR_USER}"
    )
    user_id = res.data["id"]
    # Call fractal user show
    invoke_as_superuser(f"user show {user_id}")
    debug(res.data)
    assert res.retcode == 0
    assert res.data["email"] == EMAIL_USER


def test_edit_as_user(invoke, invoke_as_superuser, caplog, new_name):
    EMAIL_USER = f"{new_name()}@example.org"
    # Register a new user
    res = invoke_as_superuser(
        f"user register {EMAIL_USER} {PWD_USER} {PROJECT_DIR_USER}"
    )
    user_id = res.data["id"]
    # Call fractal user edit
    with pytest.raises(SystemExit):
        res = invoke(
            f"user edit {user_id} "
            "--new-email email@example.org --make-verified"
        )
    debug(caplog.text)
    assert "403" in caplog.text


@pytest.mark.parametrize("new_is_superuser", [True, False])
@pytest.mark.parametrize("new_is_verified", [True, False])
@pytest.mark.parametrize("new_is_non_verified", [True, False])
def test_edit_as_superuser(
    invoke_as_superuser,
    new_is_superuser,
    new_is_verified,
    new_is_non_verified,
    new_name,
):
    EMAIL_USER = f"{new_name()}@example.org"
    # Register a new user
    res = invoke_as_superuser(
        f"user register {EMAIL_USER} {PWD_USER} {PROJECT_DIR_USER}"
    )
    assert res.retcode == 0
    user_id = res.data["id"]
    # Call fractal user edit
    NEW_EMAIL = f"{new_name()}@example.org"
    cmd = (
        f"user edit {user_id} "
        f"--new-email {NEW_EMAIL} "
        f"--new-password SOMETHING "
    )
    if new_is_superuser:
        cmd = f"{cmd} --make-superuser"
    if new_is_verified:
        cmd = f"{cmd} --make-verified"
    if new_is_non_verified:
        cmd = f"{cmd} --remove-verified"

    if new_is_verified and new_is_non_verified:
        with pytest.raises(SystemExit):
            invoke_as_superuser(cmd)
    elif new_is_verified or new_is_non_verified:
        res = invoke_as_superuser(cmd)
        assert res.retcode == 0
        assert res.data["email"] == NEW_EMAIL
        assert res.data["is_superuser"] == new_is_superuser
        assert (
            res.data["is_verified"]
            if new_is_verified
            else not res.data["is_verified"]
        )
    else:
        res = invoke_as_superuser(cmd)
        assert res.retcode == 1
        assert res.data == (
            "Cannot use `--new-email` without `--make-verified` or "
            "`--remove-verified`"
        )

    # If the user was made a superuser, check that we can go back to normal
    # user
    if new_is_superuser:
        cmd = f"user edit {user_id} --remove-superuser"
        debug(cmd)
        res = invoke_as_superuser(cmd)
        debug(res.data)
        assert res.retcode == 0
        assert not res.data["is_superuser"]

    # If the user was made verified, check that we can go back to normal
    # user
    if new_is_non_verified:
        cmd = f"user edit {user_id} --make-verified"
        debug(cmd)
        res = invoke_as_superuser(cmd)
        debug(res.data)
        assert res.retcode == 0
        assert res.data["is_verified"]


def test_edit_arguments(invoke_as_superuser):
    # Test that superuser flags are mutually exclusive
    with pytest.raises(SystemExit):
        cmd = "user edit SOME_USER_ID --make-superuser --remove-superuser"
        invoke_as_superuser(cmd)


def test_whoami_as_user(invoke, tester):
    res = invoke("user whoami")
    assert res.retcode == 0
    debug(res.data)
    assert res.data["email"] == tester["email"]
    assert not res.data["is_superuser"]
    user_id = res.data["id"]

    # Test user whoami with --batch flag
    res = invoke("--batch user whoami")
    debug(res.data)
    assert res.data == user_id
    assert res.retcode == 0


def test_whoami_as_superuser(invoke_as_superuser, superuser):
    res = invoke_as_superuser("user whoami")
    assert res.retcode == 0
    debug(res.data)
    assert res.data["email"] == superuser["email"]
    assert res.data["is_superuser"]


def test_user_set_groups(invoke_as_superuser, user_factory, new_name):
    # get default group
    res = invoke_as_superuser("group list --user-ids")
    default_group = next(
        group
        for group in res.data
        if group["name"] == FRACTAL_DEFAULT_GROUP_NAME
    )
    default_group_id = default_group["id"]
    # create 2 new users
    user1 = user_factory(email=f"{new_name()}@example.org", password="psw1")
    assert len(user1["group_ids_names"]) == 1
    user1_id = user1["id"]
    user2 = user_factory(email=f"{new_name()}@example.org", password="psw2")
    assert len(user2["group_ids_names"]) == 1
    user2_id = user2["id"]
    # create 2 new groups
    group1 = invoke_as_superuser(f"group new {new_name()}")
    assert len(group1.data["user_ids"]) == 0
    group1_id = group1.data["id"]
    group2 = invoke_as_superuser(f"group new {new_name()}")
    assert len(group2.data["user_ids"]) == 0
    group2_id = group2.data["id"]

    with pytest.raises(SystemExit):
        # no arguments
        invoke_as_superuser("user set-groups")
    with pytest.raises(SystemExit):
        # no group_ids list
        invoke_as_superuser(f"user set-groups {user1_id}")
    with pytest.raises(SystemExit):
        # group_ids must be a list of integers
        invoke_as_superuser(f"user set-groups {user1_id} {group1_id} foo")
    with pytest.raises(SystemExit):
        # there must always be the default group id in group_ids
        invoke_as_superuser(f"user set-groups {user1_id} {group1_id}")
    with pytest.raises(SystemExit):
        # repeated elements in group_ids are forbidden
        invoke_as_superuser(
            "user set-groups "
            f"{user1_id} {default_group_id} {group1_id} {group1_id}"
        )

    # Add user1 to group1
    res = invoke_as_superuser(
        f"user set-groups {user1_id} {group1_id} {default_group_id}"
    )
    assert len(res.data["group_ids_names"]) == 2
    group1 = invoke_as_superuser(f"group get {group1_id}")
    assert len(group1.data["user_ids"]) == 1

    # Add user2 to group1 and group2
    res = invoke_as_superuser(
        "user set-groups "
        f"{user2_id} {group2_id} {group1_id} {default_group_id}"
    )
    assert len(res.data["group_ids_names"]) == 3
    group1 = invoke_as_superuser(f"group get {group1_id}")
    assert len(group1.data["user_ids"]) == 2
    group2 = invoke_as_superuser(f"group get {group2_id}")
    assert len(group2.data["user_ids"]) == 1

    # Add user1 to group2 and remove them from group1
    res = invoke_as_superuser(
        f"user set-groups {user1_id} {group2_id} {default_group_id}"
    )
    assert len(res.data["group_ids_names"]) == 2
    group1 = invoke_as_superuser(f"group get {group1_id}")
    assert len(group1.data["user_ids"]) == 1
    group2 = invoke_as_superuser(f"group get {group2_id}")
    assert len(group2.data["user_ids"]) == 2


def test_edit_project_dirs(
    invoke_as_superuser,
    new_name,
):
    # Register a new user
    EMAIL_USER = f"{new_name()}@example.org"
    PROJECT_DIR_1 = "/tmp/1"
    res = invoke_as_superuser(
        f"user register {EMAIL_USER} {PWD_USER} {PROJECT_DIR_1}"
    )
    assert res.retcode == 0
    user_id = res.data["id"]
    assert res.data["project_dirs"] == [PROJECT_DIR_1]

    # Add a second project dir
    PROJECT_DIR_2 = "/tmp/2"
    cmd = f"user edit {user_id} " f"--add-project-dir {PROJECT_DIR_2}"
    res = invoke_as_superuser(cmd)
    assert res.retcode == 0
    assert res.data["project_dirs"] == [PROJECT_DIR_1, PROJECT_DIR_2]

    # Add a third project dir and remove the second one
    PROJECT_DIR_3 = "/tmp/3"
    cmd = (
        f"user edit {user_id} "
        f"--add-project-dir {PROJECT_DIR_3} "
        f"--remove-project-dir {PROJECT_DIR_2}"
    )
    res = invoke_as_superuser(cmd)
    assert res.retcode == 0
    assert res.data["project_dirs"] == [PROJECT_DIR_1, PROJECT_DIR_3]



import json
from pathlib import Path

from ..authclient import AuthClient
from ..interface import Interface
from ..response import check_response


def post_resource(
    client: AuthClient,
    *,
    json_file: str,
    batch: bool = False,
) -> Interface:
    # Prepare a ResourceCreate request body
    with Path(json_file).open("r") as f:
        payload = json.load(f)

    # Send API request
    res = client.post("admin/v2/resource/", json=payload)
    resource = check_response(res, expected_status_code=201)
    if batch:
        return Interface(retcode=0, data=resource["id"])
    else:
        return Interface(retcode=0, data=resource)



import json

import pytest  # noqa F401
from devtools import debug

TIMEOUT = 15.0


def test_workflow_new(invoke, new_name):
    PROJECT_NAME = new_name()
    res = invoke(f"project new {PROJECT_NAME}")
    debug(res)
    debug(res.data)
    proj = res.data
    assert proj["name"] == PROJECT_NAME
    project_id = proj["id"]

    WORKFLOW_NAME = new_name()
    res = invoke(f"workflow new {WORKFLOW_NAME} {project_id}")
    wf = res.data
    debug(wf)
    assert res.retcode == 0
    assert wf["name"] == WORKFLOW_NAME
    assert wf["project_id"] == project_id

    # Include --batch
    WORKFLOW_NAME = new_name()
    res = invoke(f"--batch workflow new {WORKFLOW_NAME} {project_id}")
    assert res.retcode == 0
    debug(res.data)
    assert isinstance(res.data, int)


def test_workflow_delete(invoke, new_name):
    # Create project
    res_pj = invoke(f"project new {new_name()}")
    assert res_pj.retcode == 0
    project_id = res_pj.data["id"]

    # Create workflow
    res_wf = invoke(f"workflow new {new_name()} {project_id}")
    workflow_id = res_wf.data["id"]
    assert res_wf.retcode == 0

    # List workflows
    res_list = invoke(f"workflow list {project_id}")
    assert res_list.retcode == 0
    debug(res_list.data)
    assert len(res_list.data) == 1

    # Delete workflow
    res_delete = invoke(f"workflow delete {project_id} {workflow_id}")
    assert res_delete.retcode == 0

    # List workflows
    res_list = invoke(f"workflow list {project_id}")
    assert res_list.retcode == 0
    debug(res_list.data)
    assert len(res_list.data) == 0


def test_workflow_edit(invoke, new_name):
    # Create a project
    res_pj = invoke(f"project new {new_name()}")
    assert res_pj.retcode == 0
    project_id = res_pj.data["id"]

    # Add a workflow
    res_wf = invoke(f"workflow new {new_name()} {project_id}")
    workflow_id = res_wf.data["id"]
    assert res_wf.retcode == 0

    # Fail editing with no edits
    cmd = f"workflow edit {project_id} {workflow_id}"
    with pytest.raises(SystemExit):
        res = invoke(cmd)

    # Edit workflow name
    NAME = new_name()
    cmd = f"workflow edit {project_id} {workflow_id} --new-name {NAME}"
    debug(cmd)
    res_edit = invoke(cmd)
    assert res_edit.retcode == 0
    debug(res_edit.data)

    # List workflows, and check edit
    res = invoke(f"workflow show {project_id} {workflow_id}")
    debug(res.data)
    assert res.retcode == 0
    assert res.data["name"] == NAME


def test_workflow_list(invoke, new_name):
    PROJECT_NAME = new_name()
    res_pj = invoke(f"project new {PROJECT_NAME}")
    project_id = res_pj.data["id"]
    debug(project_id)

    res_wf = invoke(f"workflow new {new_name()} {project_id}")
    res_wf.show()
    assert res_wf.retcode == 0

    res_wf = invoke(f"workflow new {new_name()} {project_id}")
    res_wf.show()
    assert res_wf.retcode == 0

    res_list = invoke(f"workflow list {project_id}")
    debug(res_list)
    debug(res_list.data)
    assert res_list.retcode == 0
    assert len(res_list.data) == 2


def test_workflow_list_when_two_projects_exist(invoke, new_name):
    res_pj1 = invoke(f"project new {new_name()}")
    res_pj2 = invoke(f"project new {new_name()}")
    project_id_1 = res_pj1.data["id"]
    project_id_2 = res_pj2.data["id"]

    NUM_WF_PROJECT_1 = 2
    NUM_WF_PROJECT_2 = 4

    for _ in range(NUM_WF_PROJECT_1):
        res_wf = invoke(f"workflow new {new_name()} {project_id_1}")
        assert res_wf.retcode == 0

    for _ in range(NUM_WF_PROJECT_2):
        res_wf = invoke(f"workflow new {new_name()} {project_id_2}")
        assert res_wf.retcode == 0

    res_list_1 = invoke(f"workflow list {project_id_1}")
    assert res_list_1.retcode == 0
    assert len(res_list_1.data) == NUM_WF_PROJECT_1

    res_list_2 = invoke(f"workflow list {project_id_2}")
    assert res_list_2.retcode == 0
    assert len(res_list_2.data) == NUM_WF_PROJECT_2


def test_workflow_add_task(
    caplog,
    invoke,
    task_factory,
    workflow_factory,
    tmp_path,
    new_name,
):
    """
    GIVEN a workflow
    WHEN
        the client is invoked to add a task, with several different options
        (custom args, custom meta, --batch)
    THEN
        the WorkflowTask's are correctly registered in the db, and the returned
        object has the right properties
    """
    res = invoke(f"project new {new_name()}")
    project_id = res.data["id"]

    wf = workflow_factory(name=new_name(), project_id=project_id)
    t = task_factory(name=new_name(), command_parallel="pwd")

    TYPE_FILTERS = {"b": True}
    ARGS_PARALLEL = {"image_dir": "/asdasd"}
    ARGS_NON_PARALLEL = {"image_dir": "/dsadsa"}
    META_PARALLEL = {"a": "b"}
    META_NON_PARALLEL = {"c": "d"}

    type_filters_file = tmp_path / "type_filters.json"
    with type_filters_file.open("w") as f:
        json.dump(TYPE_FILTERS, f)

    args_parallel_file = tmp_path / "args_parallel_file.json"
    with args_parallel_file.open("w") as f:
        json.dump(ARGS_PARALLEL, f)

    args_non_parallel_file = tmp_path / "args_non_parallel_file.json"
    with args_non_parallel_file.open("w") as f:
        json.dump(ARGS_NON_PARALLEL, f)

    meta_parallel_file = tmp_path / "meta_paral.json"
    with meta_parallel_file.open("w") as f:
        json.dump(META_PARALLEL, f)

    meta_non_parallel_file = tmp_path / "meta_non_paral.json"
    with meta_non_parallel_file.open("w") as f:
        json.dump(META_NON_PARALLEL, f)

    cmd = f"workflow add-task {project_id} {wf['id']}"
    # Test fail with no task_id nor task_name
    with pytest.raises(SystemExit):
        invoke(cmd)
    # Test fail with both task_id and task_name
    with pytest.raises(SystemExit):
        invoke(
            f"{cmd} --task-id {t['id']} --task-name {t['name']} "
            f"--args-parallel {args_parallel_file}"
        )
    # Test fail with both task_id and version
    with pytest.raises(SystemExit):
        invoke(
            f"{cmd} --task-id {t['id']} --task-version 1.2.3.4.5.6 "
            f"--args-parallel {args_parallel_file}"
        )
    assert caplog.records[-1].msg == (
        "Too many arguments: cannot provide both `task_id` and `task_version`."
    )

    cmd_args = (
        f"{cmd} --task-id {t['id']} --type-filters {type_filters_file} "
        f"--args-parallel {args_parallel_file} "
    )
    debug(cmd_args)
    # Test success
    res = invoke(cmd_args)
    debug(res.data)
    assert res.retcode == 0

    workflow_task = res.data[0]
    workflow_task_id_1 = workflow_task["id"]
    debug(workflow_task)
    assert workflow_task["type_filters"] == TYPE_FILTERS
    assert workflow_task["args_parallel"] == ARGS_PARALLEL

    # Add a WorkflowTask by Task.name with the --batch option
    cmd_batch = (
        f"--batch workflow add-task {project_id} {wf['id']} "
        f"--task-name {t['name']} --args-parallel {args_parallel_file}"
    )
    debug(cmd_batch)
    res = invoke(cmd_batch)
    assert res.retcode == 0
    debug(res.data)
    workflow_task_id_2 = int(res.data)

    # Add a WorkflowTask with meta-parallel args
    cmd_meta = (
        f"{cmd} --task-id {t['id']} --type-filters {type_filters_file} "
        f"--args-parallel {args_parallel_file} "
        f"--meta-parallel {meta_parallel_file} "
    )
    debug(cmd_meta)
    # Test success
    res = invoke(cmd_meta)
    debug(res.data)
    assert res.retcode == 0

    workflow_task = res.data[0]
    workflow_task_id_3 = workflow_task["id"]
    assert workflow_task["meta_parallel"] == META_PARALLEL
    assert workflow_task["args_parallel"] == ARGS_PARALLEL

    # Add a WorkflowTask with meta-non-parallel args
    t_non_parallel = task_factory(
        name=new_name(),
        command_non_parallel="pwd",
    )

    cmd_meta = (
        f"{cmd} --task-id {t_non_parallel['id']} "
        f"--type-filters {type_filters_file} "
        f"--args-non-parallel {args_non_parallel_file} "
        f"--meta-non-parallel {meta_non_parallel_file}"
    )
    debug(cmd_meta)
    # Test success
    res = invoke(cmd_meta)
    debug(res.data)
    assert res.retcode == 0

    workflow_task = res.data[0]
    workflow_task_id_4 = workflow_task["id"]
    assert workflow_task["meta_non_parallel"] == META_NON_PARALLEL
    assert workflow_task["args_non_parallel"] == ARGS_NON_PARALLEL

    # Check that the WorkflowTask's in Workflow.task_list have the correct IDs
    cmd = f"workflow show {project_id} {wf['id']}"
    res = invoke(cmd)
    assert res.retcode == 0
    workflow = res.data
    debug(workflow)
    list_IDs = [wftask["id"] for wftask in workflow["task_list"]]
    assert list_IDs == [
        workflow_task_id_1,
        workflow_task_id_2,
        workflow_task_id_3,
        workflow_task_id_4,
    ]


def test_workflow_add_task_by_name(
    invoke,
    task_factory,
    workflow_factory,
    tmp_path,
    new_name,
):
    """
    GIVEN a workflow and a task
    WHEN the client is invoked to add a task *by name*
    THEN the WorkflowTask is added (for a valid name) or an error is raised
    (for invalid name)
    """
    res = invoke(f"project new {new_name()}")
    project_id = res.data["id"]
    wf = workflow_factory(name=new_name(), project_id=project_id)
    task = task_factory(name=new_name(), command_parallel="parallel")
    debug(task)

    ARGS = {"image_dir": "/asdasd"}

    args_file = tmp_path / "args_file.json"
    with args_file.open("w") as f:
        json.dump(ARGS, f)

    cmd = (
        f"workflow add-task {project_id} {wf['id']} --task-name {task['name']}"
        f" --args-parallel {args_file}"
    )
    debug(cmd)
    res = invoke(cmd)
    assert res.retcode == 0
    debug(res.data)
    assert res.data[0]["task"]["id"] == task["id"]

    # Fail when adding task via a wrong name
    with pytest.raises(SystemExit):
        cmd = (
            f"workflow add-task {project_id} {wf['id']} "
            f"--task-name INVALID_NAME --args-parallel {args_file}"
        )
        debug(cmd)
        res = invoke(cmd)


def test_workflow_rm_task(
    invoke,
    task_factory,
    workflow_factory,
    tmp_path,
    new_name,
):
    # Create project, workflow and task
    res = invoke(f"project new {new_name()}")
    project_id = res.data["id"]
    wf = workflow_factory(name=new_name(), project_id=project_id)
    t = task_factory(name=new_name(), command_parallel="parallel")

    ARGS = {"image_dir": "/asdasd"}

    args_file = tmp_path / "args_file.json"
    with args_file.open("w") as f:
        json.dump(ARGS, f)

    # Add task to workflow, twice
    cmd = (
        f"workflow add-task {project_id} {wf['id']} --task-id {t['id']} "
        f"--args-parallel {args_file}"
    )
    res = invoke(cmd)
    assert res.retcode == 0
    res = invoke(cmd)
    assert res.retcode == 0
    workflow_task_id_1 = res.data[0]["id"]

    # Remove task 1 from workflow
    cmd = f"workflow rm-task {project_id} {wf['id']} {workflow_task_id_1}"
    debug(cmd)
    res = invoke(cmd)
    assert res.retcode == 0
    debug(res.data)


def test_workflow_edit_task(
    invoke, task_factory, workflow_factory, tmp_path, new_name
):
    """
    GIVEN a workflow
    WHEN the client is invoked to add a task, including custom args
    THEN
        the WorkflowTask is correctly registered in the db, including custom
        gargs
    """

    res = invoke(f"project new {new_name()}")
    project_id = res.data["id"]
    wf = workflow_factory(name=new_name(), project_id=project_id)
    t = task_factory(name=new_name(), command_parallel="parallel")

    TYPE_FILTERS = {"b": True}
    ARGS_PARALLEL = {"image_dir": "/asdasd"}
    ARGS_NON_PARALLEL = {"image_dir": "/dsadsa"}
    META_PARALLEL = {"a": "b"}
    META_NON_PARALLEL = {"c": "d"}

    type_filters_file = tmp_path / "type_filters.json"
    with type_filters_file.open("w") as f:
        json.dump(TYPE_FILTERS, f)

    args_parallel_file = tmp_path / "args_parallel_file.json"
    with args_parallel_file.open("w") as f:
        json.dump(ARGS_PARALLEL, f)

    args_non_parallel_file = tmp_path / "args_non_parallel_file.json"
    with args_non_parallel_file.open("w") as f:
        json.dump(ARGS_NON_PARALLEL, f)

    meta_parallel_file = tmp_path / "meta_paral.json"
    with meta_parallel_file.open("w") as f:
        json.dump(META_PARALLEL, f)

    meta_non_parallel_file = tmp_path / "meta_non_paral.json"
    with meta_non_parallel_file.open("w") as f:
        json.dump(META_NON_PARALLEL, f)

    # Create task, without overriding arguments
    cmd = (
        f"workflow add-task {project_id} {wf['id']} --task-id {t['id']} "
        f"--args-parallel {args_parallel_file}"
    )
    res = invoke(cmd)
    assert res.retcode == 0

    # Edit workflow task
    debug(res.data)
    workflow_task_id = res.data[0]["id"]
    cmd = (
        f"workflow edit-task {project_id} {wf['id']} {workflow_task_id} "
        f"--type-filters {type_filters_file}"
    )
    debug(cmd)
    res = invoke(cmd)
    assert res.retcode == 0
    assert res.data["type_filters"] == TYPE_FILTERS

    # Edit workflow task
    debug(res.data)
    workflow_task_id = res.data["id"]
    cmd = (
        f"workflow edit-task {project_id} {wf['id']} {workflow_task_id} "
        f"--args-parallel {args_parallel_file} "
        f"--meta-parallel {meta_parallel_file}"
    )
    debug(cmd)
    res = invoke(cmd)
    assert res.retcode == 0
    assert res.data["meta_parallel"] == META_PARALLEL
    assert res.data["args_parallel"] == ARGS_PARALLEL

    # Add a WorkflowTask with meta-non-parallel args
    t_non_parallel = task_factory(
        name=new_name(), command_non_parallel="non_parallel"
    )

    cmd = (
        f"workflow add-task {project_id} {wf['id']} "
        f"--task-id {t_non_parallel['id']} "
        f"--args-non-parallel {args_non_parallel_file}"
    )
    res = invoke(cmd)
    assert res.retcode == 0
    workflow_task_id = res.data[0]["id"]

    cmd = (
        f"workflow edit-task {project_id} {wf['id']} {workflow_task_id} "
        f"--type-filters {type_filters_file} "
        f"--args-non-parallel {args_non_parallel_file} "
        f"--meta-non-parallel {meta_non_parallel_file}"
    )
    debug(cmd)
    # Test success
    res = invoke(cmd)
    debug(res.data)
    assert res.retcode == 0

    workflow_task = res.data
    assert workflow_task["meta_non_parallel"] == META_NON_PARALLEL
    assert workflow_task["args_non_parallel"] == ARGS_NON_PARALLEL


def test_workflow_import(
    invoke,
    testdata_path,
    new_name,
    tmp_path,
    capsys,
):
    name1 = new_name()
    res = invoke(
        "task new "
        "--command-parallel pwd --command-non-parallel pwd --version 1 "
        f"{name1}"
    )
    debug(res.data)
    assert res.retcode == 0

    name2 = new_name()
    res = invoke(
        "task new "
        "--command-parallel pwd --command-non-parallel pwd --version 1 "
        f"{name2}"
    )
    debug(res.data)
    assert res.retcode == 0

    # create project
    PROJECT_NAME = new_name()
    res_pj = invoke(f"project new {PROJECT_NAME}")
    assert res_pj.retcode == 0
    project_id = res_pj.data["id"]

    # Fail due to missing --json-file argument
    with pytest.raises(SystemExit):
        invoke(f"workflow import --project-id {project_id}")
    # import workflow into project
    with (testdata_path / "import-export/workflow.json").open("r") as f:
        worfklow_to_import = json.load(f)
    worfklow_to_import["task_list"][0]["task"]["name"] = name1
    worfklow_to_import["task_list"][0]["task"]["pkg_name"] = name1
    filename = tmp_path / "workflow1"
    with filename.open("w") as f:
        json.dump(worfklow_to_import, f)
    res = invoke(
        f"workflow import --project-id {project_id} --json-file {filename}"
    )
    debug(res.data)
    assert res.retcode == 0

    imported_workflow = res.data

    # get the workflow from the server, and check that it is the same
    workflow_id = res.data["id"]
    res = invoke(f"workflow show {project_id} {workflow_id}")
    debug(res.retcode, res.data)
    assert res.retcode == 0
    assert res.data == imported_workflow

    # import workflow into project, with --batch
    with (testdata_path / "import-export/workflow_2.json").open("r") as f:
        worfklow_to_import = json.load(f)
    worfklow_to_import["task_list"][0]["task"]["name"] = name2
    worfklow_to_import["task_list"][0]["task"]["pkg_name"] = name2
    filename = tmp_path / "workflow2"
    with filename.open("w") as f:
        json.dump(worfklow_to_import, f)
    res = invoke(
        f"--batch workflow import --project-id {project_id} "
        f"--json-file {filename}"
    )
    debug(res.retcode, res.data)
    assert res.retcode == 0

    # import workflow into project, with --workflow-name
    res = invoke(
        f"workflow import --project-id {project_id} --json-file {filename} "
        " --workflow-name MyWorkflow-V2-xxx"
    )
    debug(res.data)
    assert res.retcode == 0
    assert res.data["name"] == "MyWorkflow-V2-xxx"

    worfklow_to_import["task_list"][0]["task"]["version"] = "2"
    filename = tmp_path / "workflow3"
    with filename.open("w") as f:
        json.dump(worfklow_to_import, f)
    with pytest.raises(SystemExit):
        invoke(
            f"workflow import --project-id {project_id} --json-file {filename}"
            " --workflow-name MyWorkflow-V2-yyy"
        )
    captured = capsys.readouterr()
    assert (
        f"Task '{name2}' (package '{name2}', version '2') not available. "
        "Available versions: ['1']."
    ) in captured.out


def test_workflow_export(
    invoke,
    workflow_factory,
    tmp_path,
    task_factory,
    new_name,
):
    res = invoke(f"project new {new_name()}")
    assert res.retcode == 0
    project_id = res.data["id"]

    NAME = new_name()
    wf = workflow_factory(project_id=project_id, name=NAME)
    prj_id = wf["project_id"]
    wf_id = wf["id"]
    filename = str(tmp_path / "exported_wf.json")

    task = task_factory(name=new_name(), command_parallel="pwd")
    res = invoke(f"workflow add-task {prj_id} {wf_id} --task-id {task['id']}")
    assert res.retcode == 0

    res = invoke(f"workflow export {prj_id} {wf_id} --json-file {filename}")
    debug(res.data)
    assert res.retcode == 0

    debug(res.data)
    with open(filename) as f:
        exported_wf = json.load(f)
        assert exported_wf["name"] == NAME
        assert "id" not in exported_wf
        assert "project_id" not in exported_wf
        for wftask in exported_wf["task_list"]:
            assert "id" not in wftask
            assert "task_id" not in wftask
            assert "workflow_id" not in wftask



import json
from pathlib import Path

import pytest
from devtools import debug
from fractal_client.cmd._aux_task_caching import TASKS_CACHE_FILENAME
from fractal_client.config import settings


COLLECTION_TIMEOUT = 15.0


def test_task_new(
    invoke,
    invoke_as_custom_user,
    tmp_path,
    new_name,
    user_factory,
):
    # create a new task with just positional required args
    args_path = str(tmp_path / "args.json")
    args = {"image_dir": "/asdasd"}
    with open(args_path, "w") as f:
        json.dump(args, f)

    meta_path = str(tmp_path / "meta.json")
    meta = {"a": "b"}
    with open(meta_path, "w") as f:
        json.dump(meta, f)

    TASK_NAME = new_name()
    res = invoke(
        f"task new {TASK_NAME} --command-parallel _command "
        f"--version _version --meta-parallel {meta_path} "
        f"--args-schema-parallel {args_path} "
        f"--args-schema-version 1.0.0 "
        "--private"
    )
    debug(res.data)
    assert res.retcode == 0
    assert res.data["name"] == TASK_NAME
    assert res.data["command_parallel"] == "_command"
    assert res.data["version"] == "_version"
    assert res.data["meta_parallel"] == meta
    assert res.data["args_schema_version"] == "1.0.0"
    first_task_id = int(res.data["id"])

    # Check that task is actually private
    new_user_credentials = dict(
        email=f"{new_name()}@example.org",
        password="1234",
    )
    user_factory(**new_user_credentials)
    with pytest.raises(SystemExit):
        res = invoke_as_custom_user(
            f"task show {first_task_id}",
            **new_user_credentials,
        )

    # create a new task with batch option
    TASK_NAME_2 = new_name()
    res = invoke(
        f"--batch task new {TASK_NAME_2} --command-parallel _command2"
    )
    res.show()
    assert res.retcode == 0
    assert res.data == str(first_task_id + 1)

    # create a new task with same name as before. Note that in check_response
    # we have sys.exit(1) when status code is not the expecte one
    with pytest.raises(SystemExit) as e:
        invoke(f"task new {TASK_NAME_2} --command-parallel _command2")
    assert e.value.code == 1

    # create a new task passing not existing file
    res = invoke(
        f"task new {new_name()} --command-parallel _command "
        "--meta-parallel ./foo.pdf"
    )
    assert res.retcode == 1

    metanp_path = str(tmp_path / "meta.json")
    metanp = {"a": "b"}
    with open(metanp_path, "w") as f:
        json.dump(metanp, f)
    res = invoke(
        f"task new {new_name()} --command-non-parallel _command_np "
        f"--meta-non-parallel {metanp_path} "
        f"--args-schema-non-parallel {args_path} "
    )
    assert res.data["args_schema_non_parallel"] == args


def test_task_edit(
    caplog,
    invoke,
    tmp_path,
    new_name,
):
    args_path = str(tmp_path / "args.json")
    args = {"image_dir": "/asdasd"}
    with open(args_path, "w") as f:
        json.dump(args, f)

    meta_path = str(tmp_path / "meta.json")
    meta = {"a": "b"}
    with open(meta_path, "w") as f:
        json.dump(meta, f)

    NAME = new_name()
    task = invoke(
        f"task new {NAME} --command-parallel _command "
        f"--version _version --meta-parallel {meta_path} "
        f"--args-schema-parallel {args_path} "
        f"--args-schema-version 1.0.0"
    )

    task.show()
    assert task.retcode == 0
    task_id = task.data["id"]

    # Test successful edit of string attributes
    NEW_COMMAND_PARALLEL = "run_parallel"
    res = invoke(
        f"task edit --id {task_id} "
        f"--command-parallel {NEW_COMMAND_PARALLEL}"
    )
    assert res.data["command_parallel"] == NEW_COMMAND_PARALLEL
    assert res.retcode == 0

    # Add non-parallel task and test command-non-parallel

    meta_path = str(tmp_path / "meta.json")
    meta = {"a": "b"}
    with open(meta_path, "w") as f:
        json.dump(meta, f)

    task_np = invoke(
        f"task new {new_name()} --command-non-parallel _command_np "
        f"--version 1.0.1 --meta-non-parallel {meta_path}"
    )

    NEW_COMMAND_NON_PARALLEL = "run_non_parallel"
    res = invoke(
        f"task edit --id {task_np.data['id']} "
        f"--command-non-parallel {NEW_COMMAND_NON_PARALLEL}"
    )
    assert res.data["command_non_parallel"] == NEW_COMMAND_NON_PARALLEL
    assert res.retcode == 0

    # Test fail with no task_id nor task_name
    with pytest.raises(SystemExit):
        res = invoke("task edit")
    # Test fail with both task_id and task_name
    with pytest.raises(SystemExit):
        res = invoke(f"task edit --id {task_id} --name {task.data['name']}")
    # Test fail with both task_id and task_version
    with pytest.raises(SystemExit):
        res = invoke(f"task edit --id {task_id} --version 1.2.3.4.5.6")
    assert caplog.records[-1].msg == (
        "Too many arguments: cannot provide both `id` and `version`."
    )
    # Test fail "name and wrong version"
    with pytest.raises(SystemExit):
        invoke("task delete --name INVALID_NAME --version INVALID_VERSION")

    input_types = {"input": True, "output": False}

    i_types_path = str(tmp_path / "itypes.json")
    with open(i_types_path, "w") as f:
        json.dump(input_types, f)

    output_types = {"input": False, "output": True}

    o_types_path = str(tmp_path / "otypes.json")
    with open(o_types_path, "w") as f:
        json.dump(output_types, f)

    # Test regular updates (both by id and name)
    res = invoke(f"task edit --id {task_id} --input-types {i_types_path}")
    assert res.data["input_types"] == input_types
    assert res.retcode == 0
    res = invoke(f"task edit --name {NAME} --output-types {o_types_path}")
    assert res.data["output_types"] == output_types
    assert res.retcode == 0

    # Test regular update by name, after deleting cache
    cache_dir = Path(settings.FRACTAL_CACHE_PATH)
    cache_file = cache_dir / TASKS_CACHE_FILENAME
    cache_file.unlink(missing_ok=True)

    res = invoke(f"task edit --name {NAME} --output-types {o_types_path}")
    assert res.data["output_types"] == output_types
    assert res.retcode == 0

    # Test failing invalid name
    fail_output_types = {"input": True, "output": False}

    f_o_types_path = str(tmp_path / "fotypes.json")
    with open(f_o_types_path, "w") as f:
        json.dump(fail_output_types, f)

    with pytest.raises(SystemExit):
        res = invoke(
            f"task edit --name INVALID_NAME --output-types {f_o_types_path}"
        )

    # Test regular update by name, after creating an invalid cache
    with cache_file.open("w") as f:
        json.dump([], f)

    new_output_types = {"input": False, "output": True}
    n_o_types_path = str(tmp_path / "notypes.json")
    with open(n_o_types_path, "w") as f:
        json.dump(new_output_types, f)

    res = invoke(f"task edit --name {NAME} --output-types {n_o_types_path}")
    assert res.data["output_types"] == new_output_types
    assert res.retcode == 0



import pytest
from devtools import debug
from fractal_client.cmd._aux_task_caching import _loose_version_parse
from packaging import version


# string, major, minor, micro, is_prerelease
VERSIONS = [
    ("0.10.0c0", 0, 10, 0, True),
    ("0.10.0b4", 0, 10, 0, True),
    ("0.10.0", 0, 10, 0, False),
    ("0.10.0alpha3", 0, 10, 0, True),
    ("0.10.0a2", 0, 10, 0, True),
    ("1.0.0", 1, 0, 0, False),
    ("0.10.0a0", 0, 10, 0, True),
    ("1.0.0rc4.dev7", 1, 0, 0, True),
    ("0.10.0beta5", 0, 10, 0, True),
    ("0.10.0alpha0", 0, 10, 0, True),
    ("3.2", 3, 2, 0, False),
    ("2", 2, 0, 0, False),
]

SORTED_VERSIONS = [
    "0.10.0a0",
    "0.10.0alpha0",
    "0.10.0a2",
    "0.10.0alpha3",
    "0.10.0b4",
    "0.10.0beta5",
    "0.10.0c0",
    "0.10.0",
    "1.0.0rc4.dev7",
    "1.0.0",
    "2",
    "3.2",
]


def test_version_parsing():
    for v_string, major, minor, micro, is_prerelease in VERSIONS:
        v = version.parse(v_string)
        debug(v_string, v.major, v.minor, v.micro, v.is_prerelease, v.pre)
        if major is not None:
            assert v.major == major
        if minor is not None:
            assert v.minor == minor
        if micro is not None:
            assert v.micro == micro
        if is_prerelease is not None:
            assert v.is_prerelease == is_prerelease

    with pytest.raises(version.InvalidVersion):
        version.parse("invalid")


def test_version_sorting():
    sorted_versions = sorted([v[0] for v in VERSIONS], key=version.parse)
    debug(sorted_versions)
    assert sorted_versions == SORTED_VERSIONS


def test_max_with_loose_version_parse():
    versions = ["invalid_1"] + [v[0] for v in VERSIONS] + ["invalid_2"]
    sorted_versions = sorted(versions, key=_loose_version_parse)
    debug(sorted_versions)
    assert sorted_versions == ["invalid_1", "invalid_2"] + SORTED_VERSIONS



import pytest
from fractal_client.response import check_response
from httpx import Request
from httpx import Response


def test_check_response(caplog):
    JSON = {"something": "else"}
    response = Response(status_code=200, json=JSON)
    checked_response = check_response(response)
    assert checked_response == JSON

    REQUEST_BODY = {"cache_dir": "xxxxxx"}
    RESPONSE_BODY = {
        "detail": [
            {
                "loc": [
                    "body",
                    "cache_dir",
                ],
                "msg": (
                    "String attribute 'cache_dir' must be an absolute path"
                    " (given 'xxxxxx')."
                ),
                "type": "value_error",
            },
        ],
    }
    response = Response(
        status_code=422,
        json=RESPONSE_BODY,
        request=Request("GET", "http://example.org", json=REQUEST_BODY),
    )
    caplog.clear()
    with pytest.raises(SystemExit):
        check_response(response)
    assert "Original request" in caplog.records[-4].getMessage()
    assert "Original payload" in caplog.records[-3].getMessage()
    print(caplog.records[-2].getMessage())
    assert "msg: String attribute " in caplog.records[-2].getMessage()
    assert "type: value_error" in caplog.records[-2].getMessage()
    assert "loc: ['body', 'cache_dir']" in caplog.records[-2].getMessage()
    assert "Terminating" in caplog.records[-1].getMessage()

    # Test accepted status codes
    RESPONSE_BODY = {"some": "response"}
    response = Response(
        status_code=123,
        json=RESPONSE_BODY,
        request=Request("GET", "http://example.org", json={"some": "request"}),
    )
    out = check_response(response, expected_status_code=123)
    assert out == RESPONSE_BODY
    out = check_response(response, expected_status_code=[123])
    assert out == RESPONSE_BODY



import pytest
from fractal_client.client import main
from fractal_client.interface import Interface


def test_unit_main(monkeypatch):
    """
    Run the `main` function.

    NOTE: Mocking `handle()` is necessary because there is no
    appropriate `sys.argv`.
    """
    import fractal_client.client

    monkeypatch.setattr(
        fractal_client.client,
        "handle",
        lambda: Interface(data="data", retcode=0),
    )
    with pytest.raises(SystemExit):
        main()



import json
import logging
import sys
from pathlib import Path

from ..authclient import AuthClient
from ..interface import Interface
from ..response import check_response


def template_show(client: AuthClient, *, template_id: int):
    res = client.get(f"api/v2/workflow-template/{template_id}/")
    data = check_response(res, expected_status_code=200)
    return Interface(retcode=0, data=data)


def template_new(
    client: AuthClient,
    *,
    workflow_id: int | None = None,
    json_file: str | None = None,
    name: str | None = None,
    version: int | None = None,
    user_group_id: int | None = None,
    batch: bool = False,
):
    if workflow_id is not None:
        if name is None or version is None:
            logging.error(
                "You must provide '--name' and '--version' when you create a "
                "template from a workflow."
            )
            sys.exit(1)
        payload = {
            "name": name,
            "version": version,
        }
        url = f"api/v2/workflow-template/?workflow_id={workflow_id}"
        if user_group_id is not None:
            url += f"&user_group_id={user_group_id}"
    else:  # json_file is not None
        with open(json_file, "r") as f:
            payload = json.load(f)
        if name is not None:
            payload["name"] = name
        if version is not None:
            payload["version"] = version
        url = "api/v2/workflow-template/import/"
        if user_group_id is not None:
            url += f"?user_group_id={user_group_id}"

    res = client.post(url, json=payload)
    template = check_response(res, expected_status_code=201)

    if batch:
        return Interface(retcode=0, data=template["id"])
    else:
        return Interface(retcode=0, data=template)


def template_delete(client: AuthClient, *, template_id: int):
    res = client.delete(f"api/v2/workflow-template/{template_id}/")
    check_response(res, expected_status_code=204)
    return Interface(retcode=0, data="")


def template_export(
    client: AuthClient,
    *,
    template_id: int,
    json_file: str,
) -> Interface:
    res = client.get(
        (f"api/v2/workflow-template/{template_id}/export/"),
    )
    template = check_response(res, expected_status_code=200)

    # mode="x" means "open for exclusive creation, failing if the file
    # already exists" https://docs.python.org/3/library/functions.html#open
    with Path(json_file).open(mode="x") as f:
        json.dump(template, f, indent=2)
    return Interface(
        retcode=0, data=f"Template {template_id} exported at {json_file}."
    )



import shlex
from pathlib import Path

import httpx
import pytest
from devtools import debug
from fractal_client import __VERSION__
from fractal_client.authclient import AuthClient
from fractal_client.client import _verify_authentication_branch
from fractal_client.client import handle
from fractal_client.cmd import version


def test_debug(invoke):
    res = invoke("--debug version")
    assert res.retcode == 0
    debug(res.data)


def test_version(invoke):
    iface = invoke("version")
    debug(iface.data)
    assert f"version: {__VERSION__}" in iface.data
    assert iface.retcode == 0


def test_version_connect_error():
    iface = version("http://localhost:9999")
    debug(iface.data)
    assert f"version: {__VERSION__}" in iface.data
    assert "refused" in iface.data
    assert iface.retcode == 0


def test_server_is_up():
    """
    GIVEN a testserver
    WHEN it gets called
    THEN it replies
    """
    res = httpx.get("http://localhost:8765/api/alive/")
    debug(res.json())
    assert res.status_code == 200


def test_user_whoami(tester, invoke):
    res = invoke("user whoami")
    user = res.data
    debug(user)
    assert res.retcode == 0
    assert user["email"] == tester["email"]


def test_user_override(user_factory, invoke, new_name):
    """
    GIVEN a user whose credentials differ from those of the environment
    WHEN the client is invoked with -u and -p
    THEN the credentials are overridden
    """
    EMAIL = f"{new_name()}@example.org"
    PASSWORD = "other_password"
    user_factory(email=EMAIL, password=PASSWORD)

    res = invoke(f"-u {EMAIL} -p {PASSWORD} project list")
    assert res.retcode == 0


def test_bad_credentials(invoke):
    """
    GIVEN a registered user
    WHEN wrong credentials are passed
    THEN the client returns an error
    """
    res = invoke("-u nouser@example.org -p nopassword project list")
    res.show()
    assert res.retcode != 0
    assert "BAD_CREDENTIALS" in res.data


def test_connecterror(override_settings):
    override_settings(
        FRACTAL_USER="admin@example.org",
        FRACTAL_PASSWORD="1234",
        FRACTAL_SERVER="http://localhost:12345",
    )
    res = handle(shlex.split("fractal user whoami"))
    debug(res.data)
    assert "ConnectError" in res.data
    assert "Hint: is http://localhost:12345 alive?" in res.data


def test_argparse_abbreviation(invoke_as_superuser, new_name):
    """
    Check that argparse abbreviations are disabled on at least one command.

    Refs:
    * https://github.com/fractal-analytics-platform/fractal/issues/440
    * https://docs.python.org/3/library/argparse.html#prefix-matching
    """

    # Successful invoke
    res = invoke_as_superuser(
        f"user register {new_name()}@example.org secret /project-dir "
        "--superuser"
    )
    res.show()
    assert res.retcode == 0

    # Failed (abbreviation-based) invoke
    with pytest.raises(SystemExit):
        invoke_as_superuser(
            f"user register {new_name()}@example.org secret2 /project-dir "
            "--super"
        )


def test_unit_verify_authentication_branch():
    # Valid cases
    _verify_authentication_branch(
        username="xxx",
        password="xxx",
        token_path=None,
    )
    _verify_authentication_branch(
        username=None,
        password=None,
        token_path="xxx",
    )

    # Invalid cases
    for username, password, token_path in [
        (None, None, None),
        ("xx", None, None),
        (None, "xx", None),
        ("xx", "xx", "xx"),
        ("xx", None, "xx"),
        (None, "xx", "xx"),
    ]:
        with pytest.raises(
            ValueError,
            match="Invalid authentication credentials",
        ):
            _verify_authentication_branch(
                username=username,
                password=password,
                token_path=token_path,
            )


def test_invalid_credentials(monkeypatch):
    import fractal_client.client

    monkeypatch.setattr(
        fractal_client.client.settings, "FRACTAL_USER", "some-user"
    )
    monkeypatch.setattr(
        fractal_client.client.settings, "FRACTAL_PASSWORD", None
    )
    interface = handle(shlex.split("fractal user whoami"))
    assert "Invalid authentication credentials" in interface.data
    assert interface.retcode == 1


def test_invalid_token_path():
    cmd = "fractal --token-path missingfile user whoami"
    interface = handle(shlex.split(cmd))
    interface.show()
    assert interface.retcode == 1


def test_valid_token_path(
    tmp_path: Path,
    monkeypatch,
    tester,
):
    # Get valid token
    with AuthClient(
        fractal_server="http://localhost:8765",
        username=tester["email"],
        password=tester["password"],
        token=None,
    ) as client:
        token_data = client.token
        debug(token_data)
    token_path = (tmp_path / "token").as_posix()

    import fractal_client.client

    monkeypatch.setattr(
        fractal_client.client.settings,
        "FRACTAL_SERVER",
        "http://localhost:8765",
    )

    # Use valid token
    with open(token_path, "w") as f:
        f.write(token_data)
    cmd = f"fractal --token-path {token_path} user whoami"
    interface = handle(shlex.split(cmd))
    assert interface.data["email"] == tester["email"]
    assert interface.retcode == 0

    # Use valid token, with newlines
    with open(token_path, "w") as f:
        f.write(f"\n\n{token_data}\n\n\n")
    cmd = f"fractal --token-path {token_path} user whoami"
    interface = handle(shlex.split(cmd))
    assert interface.data["email"] == tester["email"]
    assert interface.retcode == 0


def test_missing_fractal_server(monkeypatch):
    import fractal_client.client

    monkeypatch.setattr(
        fractal_client.client.settings,
        "FRACTAL_SERVER",
        None,
    )
    interface = handle(shlex.split("fractal user whoami"))
    assert "You should set the fractal-server URL" in interface.data
    assert interface.retcode == 1



import pytest
from fractal_client.cmd import dataset
from fractal_client.cmd import group
from fractal_client.cmd import job
from fractal_client.cmd import NoCommandError
from fractal_client.cmd import profile
from fractal_client.cmd import project
from fractal_client.cmd import resource
from fractal_client.cmd import task
from fractal_client.cmd import template
from fractal_client.cmd import user
from fractal_client.cmd import workflow


def test_invalid_commands(invoke):
    for arg in ["", " INVALID"]:
        for command in [
            "",
            "dataset",
            "project",
            "job",
            "workflow",
            "task",
            "user",
            "group",
            "resource",
            "profile",
            "template",
        ]:
            with pytest.raises(SystemExit):
                invoke(f"{command}{arg}")


def test_unit_invalid_subcommand():
    for _function in [
        project,
        dataset,
        task,
        workflow,
        job,
        user,
        group,
        resource,
        profile,
        template,
    ]:
        with pytest.raises(NoCommandError):
            _function(client=None, subcmd="invalid")



import json
from pathlib import Path

import pytest

FRACTAL_DEFAULT_GROUP_NAME = "All"


def test_template_commands(
    invoke,
    invoke_as_superuser,
    new_name,
    workflow_factory,
    task_factory,
    tmp_path: Path,
    capsys,
):
    res = invoke_as_superuser("group list --user-ids")
    default_group = next(
        group
        for group in res.data
        if group["name"] == FRACTAL_DEFAULT_GROUP_NAME
    )
    default_group_id = default_group["id"]

    res = invoke(f"project new {new_name()}")
    project = res.data
    workflow = workflow_factory(name=new_name(), project_id=project["id"])
    task = task_factory(name=new_name(), command_parallel="pwd", version="1")

    invoke(
        "workflow add-task "
        f"{project['id']} {workflow['id']} --task-id {task['id']}"
    )

    # Template new (from workflow_id)
    with pytest.raises(SystemExit):
        res = invoke(f"template new --workflow-id {workflow['id']}")
    TEMPLATE_NAME = new_name()
    res = invoke(
        f"template new --workflow-id {workflow['id']} "
        f"--name {TEMPLATE_NAME} --version 1"
    )
    assert res.retcode == 0
    template1_id = res.data["id"]
    res = invoke(
        f"--batch template new --workflow-id {workflow['id']} "
        f"--name {TEMPLATE_NAME} --version 2 "
        f"--user-group-id {default_group_id}"
    )
    assert res.retcode == 0
    template2_id = int(res.data)

    # Template show
    res = invoke(f"template show {template1_id}")
    assert res.retcode == 0
    assert res.data["user_group_id"] is None
    res = invoke(f"template show {template2_id}")
    assert res.retcode == 0
    assert res.data["user_group_id"] == default_group_id

    # Template delete
    res = invoke(f"template delete {template2_id}")
    assert res.retcode == 0
    with pytest.raises(SystemExit):
        invoke(f"template show {template2_id}")

    # Template export
    template_filename = tmp_path / "template.json"
    res = invoke(f"template export {template1_id} {template_filename}")
    assert res.retcode == 0

    # Template new (from JSON file)
    with pytest.raises(SystemExit):
        # Same (user, name, version) of template1
        invoke(
            f"template new --json-file {template_filename} "
            f"--user-group-id {default_group_id}"
        )
    res = invoke(
        "--batch "
        f"template new --json-file {template_filename} --name {new_name()} "
        f"--user-group-id {default_group_id}"
    )
    assert res.retcode == 0
    template3_id = int(res.data)
    res = invoke(
        "--batch "
        f"template new --json-file {template_filename} --version 2 "
        f"--user-group-id {default_group_id}"
    )
    assert res.retcode == 0

    # Import workflow from template
    with pytest.raises(SystemExit):
        # Same (project_id, name) of workflow
        invoke(f"workflow import-from-template {project['id']} {template3_id}")
    res = invoke(f"--batch workflow list {project['id']}")
    assert res.retcode == 0
    assert len(res.data.split()) == 1
    res = invoke(
        f"workflow import-from-template {project['id']} {template3_id} "
        f"--name {new_name()}"
    )
    assert res.retcode == 0
    res = invoke(f"--batch workflow list {project['id']}")
    assert res.retcode == 0
    assert len(res.data.split()) == 2
    res = invoke(
        "--batch "
        f"workflow import-from-template {project['id']} {template3_id} "
        f"--name {new_name()}"
    )
    assert res.retcode == 0
    assert len(res.data.split()) == 2
    res = invoke(f"--batch workflow list {project['id']}")
    assert res.retcode == 0
    assert len(res.data.split()) == 3

    # flexibility
    with template_filename.open("r") as f:
        template_to_import = json.load(f)
    NEW_VERSION = "1234"
    template_to_import["data"]["task_list"][0]["task"]["version"] = NEW_VERSION
    template_to_import["name"] = new_name()
    template_to_import["data"]["name"] = new_name()
    with template_filename.open("w") as f:
        json.dump(template_to_import, f)
    res = invoke(f"template new --json-file {template_filename}")
    assert res.retcode == 0
    template4_id = res.data["id"]
    with pytest.raises(SystemExit):
        invoke(
            f"workflow import-from-template {project['id']} {template4_id} "
            f"--name {new_name()}"
        )
    assert (
        f"Task '{task['name']}' "
        f"(package '{task['name']}', version '{NEW_VERSION}') not available. "
        f"Available versions: ['{task['version']}']."
    ) in capsys.readouterr().out



import json
import time
from pathlib import Path

import pytest
from devtools import debug
from httpx import Request
from httpx import Response

TIMEOUT = 25.0


def test_job_submit(
    invoke,
    project_factory,
    tmp_path: Path,
    testdata_path: Path,
    new_name,
    invoke_as_superuser,
    fractal_tasks_mock,
):
    # Collect tasks

    res = invoke(f"--batch task collect {fractal_tasks_mock} --private")
    assert res.retcode == 0
    activity_id = res.data

    # Create a project
    project = project_factory(name=new_name())
    project_id = project["id"]

    type_filters = {"a": True, "b": False}
    type_filters_file = tmp_path / "type_filters.json"
    with type_filters_file.open("w") as f:
        json.dump(type_filters, f)

    # Add a `tmp_path` to the user's `project_dirs`
    res = invoke("--batch user whoami")
    user_id = int(res.data)
    res = invoke_as_superuser(
        f"user edit {user_id} --add-project-dir {tmp_path.as_posix()}"
    )
    assert res.retcode == 0

    # Create dataset (batch)
    res = invoke(
        "--batch "
        f"project add-dataset {project_id} {new_name()} "
        f"--project-dir {tmp_path.as_posix()} --zarr-subfolder zarr"
    )
    assert res.retcode == 0
    dataset_id = int(res.data)

    # Wait for task collection to end
    starting_time = time.perf_counter()
    while True:
        res1 = invoke(f"task check-collection {activity_id}")
        if res1.data["status"] == "OK":
            debug(res1.data)
            break
        time.sleep(0.1)
        assert time.perf_counter() - starting_time < TIMEOUT

    wf_json = (testdata_path / "import-export/wf3.json").as_posix()
    res = invoke(
        f"workflow import --project-id {project_id} --json-file {wf_json}"
    )
    workflow = res.data
    workflow_id = workflow["id"]
    debug(workflow)

    FIRST_TASK_INDEX = 0
    LAST_TASK_INDEX = 0
    WORKER_INIT = "export MYVARIABLE=MYVALUE"

    res = invoke(
        f"job submit {project_id} {workflow_id} {dataset_id} "
        f"--start {FIRST_TASK_INDEX} --end {LAST_TASK_INDEX} "
        f'--worker-init "{WORKER_INIT}"'
    )
    assert res.retcode == 0
    job1 = res.data
    job1_id = job1["id"]
    assert job1["status"] == "submitted"
    assert job1["first_task_index"] == FIRST_TASK_INDEX
    assert job1["last_task_index"] == LAST_TASK_INDEX
    assert job1["worker_init"] == WORKER_INIT

    # Check that job completed successfully
    cmd = f"job show {project_id} {job1_id}"
    starting_time = time.perf_counter()
    debug(cmd)
    while True:
        res = invoke(cmd)
        job1 = res.data
        debug(job1)
        assert res.retcode == 0
        if job1["status"] == "done":
            break
        elif job1["status"] == "failed":
            raise RuntimeError(job1)
        time.sleep(0.1)
        assert time.perf_counter() - starting_time < TIMEOUT
    assert job1["log"] is not None

    # Prepare and run a workflow with a failing task
    FIRST_TASK_INDEX = 0
    LAST_TASK_INDEX = 1
    res = invoke(
        f"--batch "
        f"job submit {project_id} {workflow_id} {dataset_id} "
        f"--start {FIRST_TASK_INDEX} --end {LAST_TASK_INDEX} "
        f'--worker-init "{WORKER_INIT}"'
    )
    assert res.retcode == 0
    job2_id = res.data

    # Verify that status is failed, and that there is a log
    cmd = f"--batch job show {project_id} {job2_id}"
    starting_time = time.perf_counter()
    while True:
        res = invoke(cmd)
        status = res.data
        debug(status)
        assert res.retcode == 0
        if status == "failed":
            break
        time.sleep(0.1)
        assert time.perf_counter() - starting_time < TIMEOUT

    # Run job list with/without --batch
    res = invoke(f"--batch job list {project_id}")
    assert res.retcode == 0
    assert res.data == f"{job2_id} {job1_id}"
    res = invoke(f"job list {project_id}")
    assert res.retcode == 0
    assert {job["id"] for job in res.data} == {job1_id, job2_id}

    # Download logs / success
    log1_dir = tmp_path / "log1"
    cmd = (
        f"job download-logs {project_id} {job1_id} "
        f"--output {log1_dir.as_posix()}"
    )
    res = invoke(cmd)
    assert res.retcode == 0
    files = log1_dir.glob("*")
    assert "workflow.log" in [f.name for f in files]

    # Download logs / fail because folder already exists
    log1_dir = tmp_path / "log1"
    cmd = (
        f"job download-logs {project_id} {job1_id} "
        f"--output {log1_dir.as_posix()}"
    )
    res = invoke(cmd)
    assert res.retcode == 1

    # Download logs / fail because of invalid job_id
    cmd = f"job download-logs {project_id} 9999 --output /tmp/invalid/"
    with pytest.raises(SystemExit):
        invoke(cmd)

    # --attribute-filters-json and --type-filters-json
    attribute_filters = {"x": [1, 2], "y": ["foo", "bar"]}
    attribute_filters_file = tmp_path / "attribute_filters.json"
    with attribute_filters_file.open("w") as f:
        json.dump(attribute_filters, f)

    type_filters = {"x": True, "y": False}
    type_filters_file = tmp_path / "type_filters.json"
    with type_filters_file.open("w") as f:
        json.dump(type_filters, f)

    res = invoke(
        f"job submit {project_id} {workflow_id} {dataset_id} "
        f"--attribute-filters-json {attribute_filters_file} "
        f"--type-filters-json {type_filters_file}"
    )
    assert res.retcode == 0
    assert res.data["attribute_filters"] == attribute_filters
    assert res.data["type_filters"] == type_filters


def test_job_stop(invoke, caplog):
    with pytest.raises(SystemExit):
        invoke("job stop 123456 1234546")
    EXPECTED_MSG = (
        "Stopping a job execution is not implemented "
        "for FRACTAL_RUNNER_BACKEND=local"
    )
    assert EXPECTED_MSG in caplog.text


def test_job_logs_wrong_content_type(monkeypatch, invoke, tmp_path, caplog):
    def patched_get(*args, **kwargs):
        request = Request(method="GET", url="fake-url")
        response = Response(
            status_code=200,
            headers={"content-type": "text/html"},
            request=request,
        )
        debug(response)
        return response

    import fractal_client.cmd._job

    monkeypatch.setattr(fractal_client.cmd._job.AuthClient, "get", patched_get)

    outdir = (tmp_path / "logs").as_posix()
    caplog.clear()
    with pytest.raises(SystemExit):
        invoke(f"job download-logs 111111111 222222222 --output {outdir}")
    debug(caplog.text)
    assert "Unexpected content_type" in caplog.text



import json
from typing import Any


def get_resource_data(name: str) -> dict[str, Any]:
    return {
        "type": "local",
        "name": name,
        "jobs_local_dir": "/tmp/jobs_local_dir",
        "tasks_local_dir": "/tmp/tasks_local_dir",
        "jobs_poll_interval": 5,
        "jobs_runner_config": {},
        "tasks_python_config": {
            "default_version": "3.12",
            "versions": {"3.12": "/usr/bin/python3.12"},
        },
        "tasks_pixi_config": {},
    }


def test_new_resource_and_profile(tmp_path, invoke_as_superuser, new_name):
    resource_data_1 = get_resource_data(new_name())
    resource_data_2 = get_resource_data(new_name())
    resource_path_1 = tmp_path / "res1.json"
    resource_path_2 = tmp_path / "res2.json"
    with resource_path_1.open("w") as f:
        json.dump(resource_data_1, f)
    with resource_path_2.open("w") as f:
        json.dump(resource_data_2, f)

    res = invoke_as_superuser(f"resource new {resource_path_1.as_posix()}")
    assert res.retcode == 0
    resource_id_1 = res.data["id"]

    res = invoke_as_superuser(
        f"--batch resource new {resource_path_2.as_posix()}"
    )
    assert res.retcode == 0
    resource_id_2 = int(res.data)

    profile_data_1 = dict(resource_type="local", name=f"profile {new_name()}")
    profile_data_2 = dict(resource_type="local", name=f"profile {new_name()}")
    profile_path_1 = tmp_path / "prof1.json"
    profile_path_2 = tmp_path / "prof2.json"
    with profile_path_1.open("w") as f:
        json.dump(profile_data_1, f)
    with profile_path_2.open("w") as f:
        json.dump(profile_data_2, f)

    res = invoke_as_superuser(
        f"profile new {resource_id_1} {profile_path_1.as_posix()}"
    )
    assert res.retcode == 0
    res = invoke_as_superuser(
        f"--batch profile new {resource_id_2} {profile_path_2.as_posix()}"
    )
    int(res.data)
    assert res.retcode == 0



from devtools import debug


def test_add_dataset(
    invoke,
    new_name,
    tester,
):
    # Create project
    res = invoke(f"project new {new_name()}")
    project_id = res.data["id"]

    # Create dataset (batch)
    res = invoke(
        "--batch "
        f"project add-dataset {project_id} {new_name()} "
        f"--project-dir {tester['project_dir']} --zarr-subfolder zarr"
    )
    assert res.retcode == 0
    int(res.data)

    # Create dataset (non batch)
    res = invoke(f"project add-dataset {project_id} {new_name()}")
    assert res.retcode == 0
    assert "id" in res.data.keys()


def test_edit_dataset(invoke, tester, new_name):
    res = invoke(f"project new {new_name()}")
    project_id = res.data["id"]

    res = invoke(f"project add-dataset {project_id} {new_name()}")
    dataset_id = res.data["id"]

    NAME = new_name()

    res = invoke(f"dataset edit {project_id} {dataset_id} --new-name {NAME}")
    res.show()
    assert res.data["name"] == NAME
    assert res.retcode == 0


def test_delete_dataset(invoke, new_name, tester):
    # Create a project with its default dataset
    res = invoke(f"project new {new_name()}")
    project_id = res.data["id"]

    res = invoke(f"project add-dataset {project_id} {new_name()}")
    dataset_id = res.data["id"]

    # Delete dataset
    res = invoke(f"dataset delete {project_id} {dataset_id}")
    debug(res.data)
    # Check that dataset show fails
    res = invoke(f"dataset show {project_id} {dataset_id}")
    assert res.data["detail"] == "Dataset not found"


def test_show_dataset(invoke, new_name):
    # Create a project with its default dataset
    res = invoke(f"project new {new_name()}")
    project_id = res.data["id"]

    res = invoke(f"project add-dataset {project_id} {new_name()}")
    dataset_id = res.data["id"]

    res = invoke(f"dataset show {project_id} {dataset_id}")
    res.show()
    assert res.retcode == 0



import pytest
from devtools import debug


def test_project_create(invoke, new_name):
    PROJECT_NAME = new_name()
    res = invoke(f"project new {PROJECT_NAME}")
    debug(res)
    assert res.data["name"] == PROJECT_NAME


def test_project_delete(invoke, new_name):
    # Create project
    res = invoke(f"project new {new_name()}")
    res.show()
    project_id_1 = res.data["id"]

    # Show project
    res = invoke(f"project show {project_id_1}")
    res.show()

    # Delete project
    res = invoke(f"project delete {project_id_1}")
    assert res.retcode == 0

    # Try to show deleted project, and fail
    with pytest.raises(SystemExit):
        res = invoke(f"project show {project_id_1}")


def test_project_create_batch(invoke, new_name):
    res = invoke("project list")
    initial_projects = res.data

    res = invoke(f"--batch project new {new_name()}")
    debug(res)
    debug(res.data)
    project_id = res.data

    res = invoke("project list")
    assert len(res.data) == len(initial_projects) + 1
    assert any(project["id"] == project_id for project in res.data)


def test_project_list(invoke, new_name):
    res = invoke("project list")
    initial_projects = len(res.data)

    res = invoke(f"--batch project new {new_name()}")

    project0_id = res.data
    res = invoke("--batch " f"project add-dataset {project0_id} {new_name()}")
    res = invoke(f"--batch project new {new_name()}")

    res = invoke("project list")
    debug(res)
    res.show()
    assert len(res.data) == initial_projects + 2


@pytest.mark.parametrize("patch_name", [True, False])
def test_edit_project(invoke, new_name, patch_name: bool):
    name = new_name()
    res = invoke(f"project new {name}")
    project = res.data
    project_id = project["id"]

    cmd = f"project edit {project_id}"
    if patch_name:
        NEW_NAME = new_name()
        cmd += f" --new-name {NEW_NAME}"

    res = invoke(cmd)
    debug(res)

    assert res.retcode == 0
    new_project = res.data
    if patch_name:
        assert new_project["name"] == NEW_NAME
    else:
        assert new_project["name"] == name



import pytest

FRACTAL_DEFAULT_GROUP_NAME = "All"


def test_group_commands_auth(invoke, caplog):
    """
    Assert 'group' commands are not accessible to standard users
    """

    def _assert_403(cmd):
        caplog.clear()
        with pytest.raises(SystemExit):
            invoke(cmd)
        assert "403" in caplog.text

    _assert_403(cmd="group list")
    _assert_403(cmd="group get 1")
    _assert_403(cmd="group new foo")


def test_group_commands(
    user_factory, invoke_as_superuser, new_name, superuser
):
    # get default group id and superuser id
    res = invoke_as_superuser("group list --user-ids")
    assert res.retcode == 0
    initial_number_of_groups = len(res.data)
    default_group = next(
        group
        for group in res.data
        if group["name"] == FRACTAL_DEFAULT_GROUP_NAME
    )
    initial_number_of_users = len(default_group["user_ids"])

    default_group_id = default_group["id"]
    superuser_id = superuser["id"]

    # create 3 standard users (by default in default group)
    user1 = user_factory(email=f"{new_name()}@example.org", password="psw1")
    user1_id = user1["id"]
    assert user1["group_ids_names"] == [[default_group_id, "All"]]
    user2 = user_factory(email=f"{new_name()}@example.org", password="psw2")
    user2_id = user2["id"]
    assert user2["group_ids_names"] == [[default_group_id, "All"]]
    user3 = user_factory(email=f"{new_name()}@example.org", password="psw3")
    user3_id = user3["id"]
    assert user3["group_ids_names"] == [[default_group_id, "All"]]

    res = invoke_as_superuser("group list --user-ids")
    assert len(res.data) == initial_number_of_groups
    assert len(res.data[0]["user_ids"]) == initial_number_of_users + 3

    # Create 2 new empty groups (`group new`)

    with pytest.raises(SystemExit):
        # missing 'name'
        invoke_as_superuser("group new")

    NEW_NAME = new_name()
    res = invoke_as_superuser(f"group new {NEW_NAME}")
    assert res.retcode == 0
    assert res.data["name"] == NEW_NAME
    assert res.data["user_ids"] == []
    group1_id = res.data["id"]

    res = invoke_as_superuser(f"group new {new_name()}")
    group2_id = res.data["id"]

    # Add users to groups (`group add-user/remove-user`)

    with pytest.raises(SystemExit):
        # missing both arguments
        invoke_as_superuser("group add-user")
    with pytest.raises(SystemExit):
        # missing one argument
        invoke_as_superuser("group add-user 1")

    with pytest.raises(SystemExit):
        # user already in group
        invoke_as_superuser(
            f"group add-user {default_group_id} {superuser_id}"
        )
    with pytest.raises(SystemExit):
        # non existing user
        invoke_as_superuser(f"group add-user {default_group_id} 9999")

    # add `user1` and `user2` to `group1`
    invoke_as_superuser(f"group add-user {group1_id} {user1_id}")
    res = invoke_as_superuser(f"group add-user {group1_id} {user2_id}")
    assert res.retcode == 0
    assert res.data["id"] == group1_id
    assert res.data["user_ids"] == [user1_id, user2_id]

    # add `user3` and `user2` to `group2`
    invoke_as_superuser(f"group add-user {group2_id} {user3_id}")
    res = invoke_as_superuser(f"group add-user {group2_id} {user2_id}")
    assert res.retcode == 0
    assert res.data["id"] == group2_id
    assert set(res.data["user_ids"]) == {user3_id, user2_id}
    # add also `superuser` to `group2`
    res = invoke_as_superuser(f"group add-user {group2_id} {superuser_id}")
    assert set(res.data["user_ids"]) == {user3_id, user2_id, superuser_id}

    # Check groups are updated

    res = invoke_as_superuser("group list --user-ids")
    assert len(res.data) == initial_number_of_groups + 2
    users_default_group = next(
        g["user_ids"] for g in res.data if g["id"] == default_group_id
    )
    assert len(users_default_group) == initial_number_of_users + 3
    users_group_1 = next(
        g["user_ids"] for g in res.data if g["id"] == group1_id
    )
    assert set(users_group_1) == {user1_id, user2_id}
    users_group_2 = next(
        g["user_ids"] for g in res.data if g["id"] == group2_id
    )
    assert set(users_group_2) == {user3_id, user2_id, superuser_id}

    # Remove users from group
    res = invoke_as_superuser(f"group remove-user {group2_id} {user3_id}")
    assert set(res.data["user_ids"]) == {user2_id, superuser_id}
    res = invoke_as_superuser(f"group remove-user {group2_id} {user2_id}")
    assert set(res.data["user_ids"]) == {superuser_id}
    res = invoke_as_superuser(f"group remove-user {group2_id} {superuser_id}")
    assert set(res.data["user_ids"]) == set()

    # Test `group get` command

    with pytest.raises(SystemExit):
        # missing 'group_id'
        invoke_as_superuser("group get")

    res = invoke_as_superuser(f"group get {default_group_id}")
    assert res.retcode == 0
    assert res.data["name"] == FRACTAL_DEFAULT_GROUP_NAME
    assert len(res.data["user_ids"]) == initial_number_of_users + 3

    # Test `list` without `--user-ids`

    res = invoke_as_superuser("group list")
    for group in res.data:
        assert group["user_ids"] is None

    # Test `--batch`

    res = invoke_as_superuser("--batch group list")
    assert len(res.data.split(" ")) == initial_number_of_groups + 2

    res = invoke_as_superuser(f"--batch group new {new_name()}")
    assert isinstance(res.data, int)

    # Test `whoami --viewer-paths`
    invoke_as_superuser(f"group add-user {group1_id} {superuser_id}")
    assert "viewer_paths" not in superuser
    res = invoke_as_superuser("user whoami --viewer-paths")
    assert set(res.data.get("viewer_paths")) == set(res.data["project_dirs"])



import json
import logging
import time
from urllib.request import urlopen

import pytest
from devtools import debug

logging.getLogger("httpx").setLevel(logging.DEBUG)

NUM_MOCK_TASKS = 18


def test_task_collection_command(invoke, caplog):
    """
    Test that all `task collect` options are correctly parsed and included in
    the the payload for the API request.
    """
    INVALID_PYTHON_VERSION = "xxx"
    PACKAGE = "devtools"
    PACKAGE_VERSION = "0.11.0"
    PYTHON_VERSION = INVALID_PYTHON_VERSION
    PACKAGE_EXTRAS = "a,b,c"
    with pytest.raises(SystemExit):
        invoke(
            "task collect "
            f"{PACKAGE} "
            f"--package-version {PACKAGE_VERSION} "
            f"--python-version {PYTHON_VERSION} "
            f"--package-extras {PACKAGE_EXTRAS} "
            "--pre-pinned-dependency pydantic=1.10.0 "
            "--post-pinned-dependency pydantic=1.10.0"
        )
    debug(caplog.text)
    assert "Server returned 422" in caplog.text
    assert f"input_value='{INVALID_PYTHON_VERSION}'" in caplog.text


def test_task_collection_invalid_pinned_dependency(invoke, caplog):
    """
    Test the case where `pinned_package_versions_pre` has the wrong format.
    """
    PACKAGE = "devtools"
    with pytest.raises(SystemExit):
        invoke(
            f"task collect {PACKAGE} --pre-pinned-dependency invalid-string"
        )
    # Check that payload was prepared correctly
    error_line = next(
        record.message
        for record in caplog.records
        if "Invalid pin:" in record.message
    )
    debug(error_line)
    assert error_line is not None


def test_task_collection(
    invoke_as_custom_user, user_factory, new_name, fractal_tasks_mock
):
    """
    GIVEN a pip installable package containing fractal-compatible tasks
    WHEN the collection subcommand is called
    THEN
        * the collection is initiated in the background
        * the server returns immediately
    """
    COLLECTION_TIMEOUT = 15.0

    new_user = dict(email=f"{new_name()}@example.org", password="1234")
    user_factory(**new_user)

    res = invoke_as_custom_user("task list", **new_user)
    initial_task_list = len(res.data)

    res0 = invoke_as_custom_user(
        f"task collect --private {fractal_tasks_mock}",
        **new_user,
    )
    debug(res0.data)
    activity_id = res0.data["id"]

    # Wait until collection is complete
    starting_time = time.perf_counter()
    while True:
        res1 = invoke_as_custom_user(
            f"task check-collection {activity_id}", **new_user
        )
        assert res1.retcode == 0
        time.sleep(0.1)
        if res1.data["status"] == "OK":
            debug(res1.data)
            break
        assert time.perf_counter() - starting_time < COLLECTION_TIMEOUT

    # Check successful status and no logs
    res2 = invoke_as_custom_user(
        f"task check-collection {activity_id}", **new_user
    )
    assert res2.retcode == 0
    assert res2.data["status"] == "OK"
    assert res2.data["log"] is None

    # Check logs
    res3 = invoke_as_custom_user(
        f"task check-collection {activity_id} --include-logs", **new_user
    )
    assert res3.retcode == 0
    assert res3.data["status"] == "OK"
    assert res3.data["log"] is not None

    # Check task list
    res = invoke_as_custom_user("task list", **new_user)
    assert len(res.data) == initial_task_list + NUM_MOCK_TASKS

    # Second collection
    with pytest.raises(SystemExit):
        invoke_as_custom_user(f"task collect {fractal_tasks_mock}", **new_user)


def test_task_collection_custom(
    user_factory, new_name, tmp_path, invoke_as_custom_user, caplog
):
    new_user = dict(email=f"{new_name()}@example.org", password="1234")
    user_factory(**new_user)

    python_interpreter = "/usr/local/bin/python"
    package_name = "pip"
    manifest = str(tmp_path / "manifest.json")

    # Download and write a valid Manifest
    manifest_url = (
        "https://github.com/fractal-analytics-platform/fractal-server/"
        "raw/main/tests/v2/fractal_tasks_mock/src/fractal_tasks_mock/"
        "__FRACTAL_MANIFEST__.json"
    )
    with urlopen(manifest_url) as f:
        manifest_dict = json.loads(f.read())
    with open(manifest, "w") as f:
        json.dump(manifest_dict, f)

    cmd = (
        f"task collect-custom --private --package-name {package_name} "
        f"label {python_interpreter} {manifest}"
    )
    res = invoke_as_custom_user(cmd, **new_user)
    debug(res.data)
    assert res.retcode == 0
    assert isinstance(res.data, list)

    # Second API call fails (tasks with the same identity already exist)
    caplog.clear()
    with pytest.raises(SystemExit):
        res = invoke_as_custom_user(cmd, **new_user)
    # Manifest was redacted, when logging the payload
    assert '"manifest": "[value too long - redacted]"' in caplog.text

    # Missing manifest file
    cmd = (
        f"task collect-custom --package-name {package_name} "
        f"label {python_interpreter} /foo/bar"
    )
    res = invoke_as_custom_user(cmd, **new_user)
    assert res.retcode == 1
    assert "file must be on the same machine" in res.data

    cmd = (
        "--batch task collect-custom --private --package-root /tmp --version 2"
        f" label2 {python_interpreter} {manifest}"
    )
    res = invoke_as_custom_user(cmd, **new_user)
    assert res.retcode == 0
    assert isinstance(res.data, str)

    # test that '--package-root' and '--package-name' are mutually exclusive
    cmd = (
        "task collect-custom --private"
        f"--package-root /tmp --package-name {package_name} "
        f"label3 {python_interpreter} {manifest}"
    )
    with pytest.raises(SystemExit):
        res = invoke_as_custom_user(cmd, **new_user)



import pytest
from fractal_client.cmd._aux_task_caching import _get_matching_tasks
from fractal_client.cmd._aux_task_caching import _search_in_task_list
from fractal_client.cmd._aux_task_caching import FractalCacheError


def test_get_matching_tasks():
    """Test all possible cases for function `_get_matching_tasks`"""

    a = dict(name="bob", id=1, version="0.1.1")
    b = dict(name="bob", id=2, version="1.2.0")
    c = dict(name="bob", id=3, version="1.3.1")
    d = dict(name="bar", id=4, version="0.0.0")
    e = dict(name="bar", id=5, version=None)

    TASK_LIST = [a, b, c, d, e]

    res = _get_matching_tasks(TASK_LIST, name="alice")
    assert res == []

    res = _get_matching_tasks(TASK_LIST, name="bob")
    assert res == [a, b, c]

    res = _get_matching_tasks(TASK_LIST, name="bob", version="1.2.0")
    assert res == [b]

    res = _get_matching_tasks(TASK_LIST, name="bar")
    assert res == [d, e]

    res = _get_matching_tasks(TASK_LIST, name="bar", version="3.1.4")
    assert res == []


def test_search_in_task_list():
    """Test all possible cases for function `_search_in_task_list`"""

    TASK_LIST = [
        dict(name="dummy1", id=101, version="1.0.1"),
        dict(name="dummy2", id=201, version=None),
        dict(name="dummy2", id=202, version="2.0.0"),
        dict(name="dummy3", id=301, version="3.0.0"),
        dict(name="dummy3", id=302, version="3.1.4"),
        dict(name="dummy4", id=401, version="4.0.0"),
        dict(name="dummy4", id=402, version="4.1.1"),
        dict(name="dummy4", id=401, version="4.1.1"),
    ]

    # TEST zero matching

    # case 1
    with pytest.raises(FractalCacheError) as err:
        res = _search_in_task_list(task_list=TASK_LIST, name="dummy0")
    print(err.value.args[0])
    assert 'There is no task with name "dummy0"' in err.value.args[0]  # noqa

    # case 2
    with pytest.raises(FractalCacheError) as err:
        res = _search_in_task_list(
            task_list=TASK_LIST, name="dummy1", version="3.1.4"
        )
    print(err.value.args[0])
    assert (
        'There is no task with (name, version)=("dummy1", 3.1.4)'
        in err.value.args[0]
    )  # noqa

    # TEST one matching
    # case 1
    res = _search_in_task_list(task_list=TASK_LIST, name="dummy1")
    assert res == 101
    # case 2
    res = _search_in_task_list(
        task_list=TASK_LIST, name="dummy1", version="1.0.1"
    )
    assert res == 101

    # TEST multiple matching
    # case 1
    with pytest.raises(FractalCacheError) as err:
        res = _search_in_task_list(task_list=TASK_LIST, name="dummy2")
    print(err.value.args[0])
    assert "Cannot determine the latest version" in err.value.args[0]
    # case 2
    res = _search_in_task_list(task_list=TASK_LIST, name="dummy3")
    assert res == 302
    # case 3
    with pytest.raises(FractalCacheError) as err:
        res = _search_in_task_list(task_list=TASK_LIST, name="dummy4")
    print(err.value.args[0])
    assert "Multiple tasks with latest version (4.1.1)" in err.value.args[0]
    print(err.value.args[0])
    # case 4
    with pytest.raises(FractalCacheError) as err:
        res = _search_in_task_list(
            task_list=TASK_LIST, name="dummy4", version="4.1.1"
        )
    print(err.value.args[0])
    assert "Multiple tasks with version 4.1.1" in err.value.args[0]



from pathlib import Path

from devtools import debug
from fractal_client.config import settings


def test_clear_cache():
    cache_dir = settings.FRACTAL_CACHE_PATH
    debug(cache_dir)
    assert cache_dir != str(Path.home() / ".cache/fractal")


def test_override_settings(override_settings):
    override_settings(FRACTAL_CACHE_PATH="/tmp/xy")
    cache_dir = settings.FRACTAL_CACHE_PATH
    assert cache_dir == "/tmp/xy"
